/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file filter.c The stand-alone filter program. */

#include "para.h"

#include "filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "filter.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "error.h"

/** The list of all status items used by para_{server,audiod,gui}. */
const char *status_item_list[] = {STATUS_ITEM_ARRAY};

char *stat_item_values[NUM_STAT_ITEMS] = {NULL};

/** Initialize the array of errors for para_filter. */
INIT_FILTER_ERRLISTS;

/** The task that reads from stdin. */
static struct stdin_task stdin_task_struct;
/** pointer to the stdin task. */
static struct stdin_task *sit = &stdin_task_struct;

/** The task that filters the data. */
static struct filter_chain filter_chain_struct;
/** Pointer to the filter chain. */
static struct filter_chain *fc = &filter_chain_struct;

/** The task that writes converted data to stdout. */
static struct stdout_task stdout_task_struct;
/** Pointer to the stdout task. */
static struct stdout_task *sot = &stdout_task_struct;

/** Gengetopt struct that holds the command line args. */
static struct filter_args_info conf;

static int loglevel;
INIT_STDERR_LOGGING(loglevel);

static void open_filters(void)
{
	int i;
	struct filter_node *fn;

	FOR_EACH_FILTER_NODE(fn, fc, i) {
		struct filter *f = filters + fn->filter_num;
		f->open(fn);
		PARA_INFO_LOG("opened %s filter\n", f->name);
		fc->outbuf = fn->buf;
		fc->out_loaded = &fn->loaded;
	}
}

static int init_filter_chain(void)
{
	int i, ret;
	struct filter_node *fn;

	if (!conf.filter_given)
		return -E_NO_FILTERS;
	fc->num_filters = conf.filter_given;
	fc->filter_nodes = para_malloc(fc->num_filters * sizeof(struct filter_node));
	fc->inbuf = sit->buf;
	fc->in_loaded = &sit->loaded;
	fc->input_error = &sit->task.error;
	fc->task.error = 0;
	fc->output_error = &sot->task.error;
	fc->task.pre_select = filter_pre_select;
	sprintf(fc->task.status, "filter chain");

	FOR_EACH_FILTER_NODE(fn, fc, i) {
		char *fa = conf.filter_arg[i];
		fn = fc->filter_nodes + i;
		ret = check_filter_arg(fa, &fn->conf);
		if (ret < 0)
			goto err;
		fn->filter_num = ret;
		fn->fc = fc;
		INIT_LIST_HEAD(&fn->callbacks);
		PARA_DEBUG_LOG("filter #%d: %s\n", i, filters[fn->filter_num].name);
	}
	open_filters();
	return 1;
err:
	free(fc->filter_nodes);
	return ret;
}

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? filter_args_info_detailed_help
		: filter_args_info_help;

	printf_or_die("%s\n\n", FILTER_CMDLINE_PARSER_PACKAGE "-"
		FILTER_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", filter_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_filter_helps(d);
	exit(0);
}

static int parse_config(int argc, char *argv[])
{
	static char *cf; /* config file */
	struct stat statbuf;

	if (filter_cmdline_parser(argc, argv, &conf))
		return -E_FILTER_SYNTAX;
	HANDLE_VERSION_FLAG("filter", conf);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	if (!cf) {
		char *home = para_homedir();
		cf = make_message("%s/.paraslash/filter.conf", home);
		free(home);
	}
	if (!stat(cf, &statbuf)) {
		struct filter_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0
		};
		if (filter_cmdline_parser_config_file(cf, &conf, &params))
			return -E_FILTER_SYNTAX;
	}
	return 1;
}

/**
 * The main function of para_filter.
 *
 * Para_filter reads data from stdin, converts it by using a chain
 * of filters (specified on the command line) and writes the resulting
 * data to stdout.
 *
 * \param argc Number of command line options.
 * \param argv Vector of arguments.
 *
 * \return \a EXIT_SUCCESS on success, EXIT_FAILURE on errors.
 */
int main(int argc, char *argv[])
{
	int ret;
	static struct sched s;

	stdin_set_defaults(sit);
	sit->buf = para_malloc(sit->bufsize),

	filter_init();
	ret = parse_config(argc, argv);
	if (ret < 0)
		goto out;
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	ret = init_filter_chain();
	if (ret < 0)
		goto out;
	sit->output_error = &fc->task.error;

	stdout_set_defaults(sot);
	sot->buf = fc->outbuf;
	sot->loaded = fc->out_loaded;
	sot->input_error = &fc->task.error;

	register_task(&sit->task);
	register_task(&fc->task);
	register_task(&sot->task);
	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	ret = schedule(&s);
	close_filters(fc);
out:
	free(sit->buf);
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
