/*
 * Copyright (C) 1997-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file para.h global paraslash definitions */

#include "config.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h> /* time(), localtime() */
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h> /* needed by create_pf_socket */
#include <string.h>
#include <assert.h>
#include "gcc-compat.h"

/** used in various contexts */
#define MAXLINE 255

/** Compute the minimum of \a x and \a y. */
#define PARA_MIN(x, y) ({ \
	typeof(x) _min1 = (x); \
	typeof(y) _min2 = (y); \
	(void) (&_min1 == &_min2); \
	_min1 < _min2 ? _min1 : _min2; })

/** Compute the maximum of \a x and \a y. */
#define PARA_MAX(x, y) ({ \
	typeof(x) _max1 = (x); \
	typeof(y) _max2 = (y); \
	(void) (&_max1 == &_max2); \
	_max1 < _max2 ? _max2 : _max1; })

/** Compute the absolute value of \a x. */
#define PARA_ABS(x) ({ \
	typeof(x) _x = (x); \
	_x > 0? _x : -_x; })

/** Debug loglevel, gets really noisy. */
#define LL_DEBUG 0
/** Still noisy, but won't fill your disk. */
#define LL_INFO  1
/** Normal, but significant event. */
#define LL_NOTICE 2
/** Unexpected event that can be handled. */
#define LL_WARNING 3
/** Unhandled error condition. */
#define LL_ERROR 4
/** System might be unreliable. */
#define LL_CRIT 5
/** Last message before exit. */
#define LL_EMERG 6
/** Number of all loglevels. */
#define NUM_LOGLEVELS 7

/** Log messages with lower priority than that will not be compiled in. */
#define COMPILE_TIME_LOGLEVEL 0

/** \cond */
#if LL_DEBUG >= COMPILE_TIME_LOGLEVEL
#define PARA_DEBUG_LOG(f,...) para_log(LL_DEBUG, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define PARA_DEBUG_LOG(...) do {;} while (0)
#endif

#if LL_INFO >= COMPILE_TIME_LOGLEVEL
#define PARA_INFO_LOG(f,...) para_log(LL_INFO, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define PARA_INFO_LOG(...) do {;} while (0)
#endif

#if LL_NOTICE >= COMPILE_TIME_LOGLEVEL
#define PARA_NOTICE_LOG(f,...) para_log(LL_NOTICE, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define PARA_NOTICE_LOG(...) do {;} while (0)
#endif

#if LL_WARNING >= COMPILE_TIME_LOGLEVEL
#define PARA_WARNING_LOG(f,...) para_log(LL_WARNING, "%s: " f, __FUNCTION__, ##  __VA_ARGS__)
#else
#define PARA_WARNING_LOG(...) do {;} while (0)
#endif

#if LL_ERROR >= COMPILE_TIME_LOGLEVEL
#define PARA_ERROR_LOG(f,...) para_log(LL_ERROR, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define PARA_ERROR_LOG(...) do {;} while (0)
#endif

#if LL_CRIT >= COMPILE_TIME_LOGLEVEL
#define PARA_CRIT_LOG(f,...) para_log(LL_CRIT, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define PARA_CRIT_LOG(...) do {;} while (0)
#endif

#if LL_EMERG >= COMPILE_TIME_LOGLEVEL
#define PARA_EMERG_LOG(f,...) para_log(LL_EMERG, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define PARA_EMERG_LOG(...)
#endif
/** \endcond */

/**
 * define a standard log function that always writes to stderr
 *
 * \param loglevel_barrier If the loglevel of the current message
 * is less than that, the message is going to be ignored.
 *
 */
#define INIT_STDERR_LOGGING(loglevel_barrier) \
	__printf_2_3 void para_log(int ll, const char* fmt,...) \
	{ \
		va_list argp; \
		if (ll < loglevel_barrier) \
			return; \
		va_start(argp, fmt); \
		vfprintf(stderr, fmt, argp); \
		va_end(argp); \
	}

/** version text used by various commands if -V switch was given */
#define VERSION_TEXT(prefix) "para_" prefix " " PACKAGE_VERSION " (" CODENAME ")" "\n" \
	"Copyright (C) 2009 Andre Noll\n" \
	"This is free software with ABSOLUTELY NO WARRANTY." \
	" See COPYING for details.\n" \
	"Written by Andre Noll.\n" \
	"Report bugs to <maan@systemlinux.org>.\n"

/** print out \p VERSION_TEXT and exit if version flag was given */
#define HANDLE_VERSION_FLAG(_prefix, _args_info_struct) \
	if (_args_info_struct.version_given) { \
		printf("%s", VERSION_TEXT(_prefix)); \
		exit(EXIT_SUCCESS); \
	}
/** sent by para_server for commands that expect a data file */
#define AWAITING_DATA_MSG "\nAwaiting Data."
/** sent by para_server if authentication was successful */
#define PROCEED_MSG "\nProceed.\n"
/** length of the \p PROCEED_MSG string */
#define PROCEED_MSG_LEN strlen(PROCEED_MSG)
/** sent by para_client to indicate the end of the command line */
#define EOC_MSG "\nEnd of Command."
/** sent by para_client, followed by the decrypted challenge number */
#define CHALLENGE_RESPONSE_MSG "challenge_response:"

/* exec */
int para_exec_cmdline_pid(pid_t *pid, const char *cmdline, int *fds);

/* time */
int tv_diff(const struct timeval *b, const struct timeval *a, struct timeval *diff);
long unsigned tv2ms(const struct timeval*);
void d2tv(double, struct timeval*);
void tv_add(const struct timeval*, const struct timeval *, struct timeval *);
void tv_scale(const unsigned long, const struct timeval *, struct timeval *);
void tv_divide(const unsigned long divisor, const struct timeval *tv,
	struct timeval *result);
int tv_convex_combination(const long a, const struct timeval *tv1,
		const long b, const struct timeval *tv2,
		struct timeval *result);
void ms2tv(const long unsigned n, struct timeval *tv);
void compute_chunk_time(long unsigned chunk_num,
		struct timeval *chunk_tv, struct timeval *stream_start,
		struct timeval *result);

/** The enum of all status items. */
enum status_items {STATUS_ITEM_ENUM NUM_STAT_ITEMS};
extern const char *status_item_list[];
/** Loop over each status item. */
#define FOR_EACH_STATUS_ITEM(i) for (i = 0; i < NUM_STAT_ITEMS; i++)
int stat_item_valid(const char *item);
int stat_line_valid(const char *);
void stat_client_write(const char *msg, int itemnum);
int stat_client_add(int fd, long unsigned mask);

__printf_2_3 void para_log(int, const char*, ...);

/**
 * Write a log message to a dynamically allocated string.
 *
 * \param fmt Usual format string.
 * \param p Result pointer.
 *
 * \sa printf(3). */
#define PARA_VSPRINTF(fmt, p) \
{ \
	int n; \
	size_t size = 100; \
	p = para_malloc(size); \
	while (1) { \
		va_list ap; \
		/* Try to print in the allocated space. */ \
		va_start(ap, fmt); \
		n = vsnprintf(p, size, fmt, ap); \
		va_end(ap); \
		/* If that worked, return the string. */ \
		if (n > -1 && n < size) \
			break; \
		/* Else try again with more space. */ \
		if (n > -1) /* glibc 2.1 */ \
			size = n + 1; /* precisely what is needed */ \
		else /* glibc 2.0 */ \
			size *= 2; /* twice the old size */ \
		p = para_realloc(p, size); \
	} \
}

/**
 *  Return a random non-negative integer in an interval.
 *
 * \param max Determines maximal possible return value.
 *
 * \return An integer between zero and \p max - 1, inclusively.
 */
static inline long int para_random(unsigned max)
{
	return ((max + 0.0) * (random() / (RAND_MAX + 1.0)));
}

/** Round up x to a multiple of y */
#define ROUND_UP(x, y) ((((x) + ((y) - 1)) / (y)) * (y))


/** Get the size of an array */
#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

/**
 * Wrapper for isspace.
 * NetBSD needs this.
 */
/*
 * The values should be cast to an unsigned char first, then to int.
 * Why? Because the isdigit (as do all other is/to functions/macros)
 * expect a number from 0 upto and including 255 as their (int) argument.
 * Because char is signed on most systems, casting it to int immediately
 * gives the functions an argument between -128 and 127 (inclusive),
 * which they will use as an array index, and which will thus fail
 * horribly for characters which have their most significant bit set.
 */
#define para_isspace(c) isspace((int)(unsigned char)(c))

/** Data that indicates an eof-condition for a fec-encoded stream. */
#define FEC_EOF_PACKET "\xec\x0d\xcc\xfe\0\0\0\0" \
	"\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0"
#define FEC_EOF_PACKET_LEN 32
